/*
 * Project: JobMarket
 * Description: asynchronous Server/Controller/Client-approach to parallel computing, based on boost::asio
 * Copyright (C) 2012 Frederik Heber. All rights reserved.
 * 
 */

/*
 * \file PoolWorker.cpp
 *
 * This file strongly follows the Serialization example from the boost::asio
 * library (see client.cpp).
 *
 *  Created on: Feb 28, 2012
 *      Author: heber
 */

// include config.h
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// boost asio needs specific operator new
#include <boost/asio.hpp>

#include "CodePatterns/MemDebug.hpp"

#include <boost/bind.hpp>
#include <boost/lexical_cast.hpp>
#include <iostream>
#include <vector>
#include "JobMarket/Connection.hpp" // Must come before boost/serialization headers.
#include <boost/serialization/vector.hpp>
#include "CodePatterns/Info.hpp"
#include "CodePatterns/Log.hpp"
#include "JobMarket/Jobs/FragmentJob.hpp"
#include "Operations/Workers/EnrollInPoolOperation.hpp"
#include "Operations/Workers/RemoveFromPoolOperation.hpp"
#include "Operations/Workers/SubmitResultOperation.hpp"
#include "JobMarket/Results/FragmentResult.hpp"
#include "JobMarket/Pool/PoolWorker.hpp"

/** Constructor for class PoolWorker.
 *
 * We automatically connect to the given pool and enroll.
 *
 * @param _io_service io service for creating connections
 * @param _host host part of MyAddress of pool to connect to
 * @param _service service part of MyAddress of pool to connect to
 * @param listenhost host part of MyAddress of this instance for listening to pool connections
 * @param listenservice seervice part of MyAddress of this instance for listening to pool connections
 */
PoolWorker::PoolWorker(
    boost::asio::io_service& _io_service,
    const std::string& _host,
    const std::string& _service,
    const std::string& listenhost,
    const std::string& listenservice) :
  io_service(_io_service),
  PoolListener(_io_service, boost::lexical_cast<unsigned short>(listenservice), *this),
  MyAddress(listenhost, listenservice),
  ServerAddress(_host, _service),
  connection_(_io_service),
  failed(boost::bind(&ExitflagContainer::setExitflag, this, ExitflagContainer::ErrorFlag)),
  OpQueue(boost::bind(&PoolWorker::finish, boost::ref(*this))),
  thread_remover(boost::bind(&PoolWorker::PoolListener_t::handle_RemoveThread, boost::ref(PoolListener)))
{
  DEBUG_FUNCTION_ENTRYEXIT

  if (PoolListener.getExitflag() == Listener::OkFlag) {
    // always enroll and make listenining initiation depend on its success
    const boost::function<void ()> initiateme =
        boost::bind(&PoolListener_t::initiateSocket, boost::ref(PoolListener));
    const boost::function<void ()> shutdown =
        boost::bind(&PoolWorker::finish, this); // no need to remove from pool
    AsyncOperation *enrollOp = new EnrollInPoolOperation(connection_, MyAddress, initiateme, shutdown);
    LOG(2, "DEBUG: Putting enroll in pool operation into queue ...");
    OpQueue.push_back(enrollOp, ServerAddress);
  } else {
    setExitflag( ErrorFlag );
    ELOG(0, "Not enrolling, we just exit due to failed listen bind.");
  }
}

/// Handle completion of a accept server operation.
void PoolWorker::PoolListener_t::handle_Accept(const boost::system::error_code& e, connection_ptr conn)
{
  DEBUG_FUNCTION_ENTRYEXIT
  if (!e)
  {
    conn->async_read(choice,
      boost::bind(&PoolWorker::PoolListener_t::handle_ReadChoice, this,
      boost::asio::placeholders::error, conn));
    // and listen for following connections
    initiateSocket();
  }
  else
  {
    // An error occurred. Log it and return. Since we are not starting a new
    // accept operation the io_service will run out of work to do and the
    // server will exit.
    setExitflag( ErrorFlag );
    ELOG(0, e.message());
  }
}

/// Handle completion of a accept server operation.
void PoolWorker::PoolListener_t::handle_ReadChoice(const boost::system::error_code& e, connection_ptr conn)
{
  DEBUG_FUNCTION_ENTRYEXIT
  if (!e)
  {
    LOG(1, "INFO: Received request for operation " << choice << ".");
    // switch over the desired choice read previously
    bool shutdownflag = false;
    switch(choice) {
      case NoServerOperation:
      {
        ELOG(1, "PoolListener_t::handle_ReadChoice() - called with NoOperation.");
        break;
      }
      case CheckAlive:
      {
        // send address as pong to ping (check alive signal)
        conn->async_write(callback.MyAddress,
            boost::bind(&PoolWorker::PoolListener_t::handle_SendAddress, this,
            boost::asio::placeholders::error, conn));
        break;
      }
      case ReceiveJob:
      {
    	// new jobs only allowed when there is no thread working
        if (workerthread == NULL) {
          JobAcceptFlag = true;
          conn->async_write(JobAcceptFlag,
              boost::bind(&PoolWorker::PoolListener_t::handle_AcceptJob, this,
              boost::asio::placeholders::error, conn));
        } else {
          JobAcceptFlag = false;
          conn->async_write(JobAcceptFlag,
              boost::bind(&PoolWorker::PoolListener_t::handle_RejectJob, this,
              boost::asio::placeholders::error, conn));
        }
        break;
      }
      case ShutdownWorker:
      {
        shutdownflag = true;
        callback.shutdown();
        break;
      }
    }
    // and listen for following connections if not to shutdown
    if (!shutdownflag)
      initiateSocket();
  }
  else
  {
    // An error occurred. Log it and return. Since we are not starting a new
    // accept operation the io_service will run out of work to do and the
    // server will exit.
    setExitflag( ErrorFlag );
    ELOG(0, e.message());
  }
}

/// Controller callback function when job has been sent.
void PoolWorker::PoolListener_t::handle_ReceiveJob(const boost::system::error_code& e, connection_ptr conn)
{
  DEBUG_FUNCTION_ENTRYEXIT

  if (!e)
  {
    ASSERT (job->getId() != JobId::NoJob,
        "PoolListener_t::handle_ReceiveJob() - received job with NoJob id.");
    LOG(1, "INFO: Working on job " << job->getId() << ".");
    boost::function< void() > workerfct = boost::bind(&PoolWorker::WorkOnJob, boost::ref(callback), boost::ref(job));
    workerthread.reset(new boost::thread(workerfct));
  }
  else
  {
    // An error occurred. Log it and return. Since we are not starting a new
    // accept operation the io_service will run out of work to do and the
    // server will exit.
    setExitflag( ErrorFlag );
    ELOG(0, e.message());
  }
}

/// Controller callback function when new job is accepted.
void PoolWorker::PoolListener_t::handle_AcceptJob(const boost::system::error_code& e, connection_ptr conn)
{
  DEBUG_FUNCTION_ENTRYEXIT

  if (!e)
  {
    ASSERT( JobAcceptFlag,
          "PoolListener_t::handle_Accept() - accepting job but flag says otherwise.");
    LOG(2, "INFO: Accepting new job.");
    conn->async_read(job,
      boost::bind(&PoolWorker::PoolListener_t::handle_ReceiveJob, this,
      boost::asio::placeholders::error, conn));
  }
  else
  {
	// An error occurred. Log it and return. Since we are not starting a new
	// accept operation the io_service will run out of work to do and the
	// server will exit.
	setExitflag( ErrorFlag );
	ELOG(0, e.message());
  }
}

/// Controller callback function when new job cannot be accepted.
void PoolWorker::PoolListener_t::handle_RejectJob(const boost::system::error_code& e, connection_ptr conn)
{
  DEBUG_FUNCTION_ENTRYEXIT

  if (!e)
  {
    ASSERT( !JobAcceptFlag,
      "PoolListener_t::handle_Accept() - not accepting job but flag says otherwise.");
  	LOG(2, "INFO: NOT accepting new job, since we are still working on one.");
  }
  else
  {
    // An error occurred. Log it and return. Since we are not starting a new
    // accept operation the io_service will run out of work to do and the
    // server will exit.
    setExitflag( ErrorFlag );
    ELOG(0, e.message());
  }
}

/// Controller callback function when address has been sent
void PoolWorker::PoolListener_t::handle_SendAddress(const boost::system::error_code& e, connection_ptr conn)
{
  DEBUG_FUNCTION_ENTRYEXIT

  if (!e)
  {
    LOG(1, "INFO: Sent address " << callback.MyAddress << " as alive confirmation.");
  }
  else
  {
    // An error occurred. Log it and return. Since we are not starting a new
    // accept operation the io_service will run out of work to do and the
    // server will exit.
    setExitflag( ErrorFlag );
    ELOG(0, e.message());
  }
}

void PoolWorker::PoolListener_t::handle_RemoveThread()
{
  // simply delete the thread
  if (workerthread != NULL)
    workerthread.reset();
}

void PoolWorker::PoolListener_t::closeSocket()
{
  if (workerthread != NULL) {
    workerthread->join();
    workerthread.reset();
  }
  // relay to base class
  Listener::closeSocket();
}

/** Works on the given job and send the results.
 *
 * @param job job to work on
 */
void PoolWorker::WorkOnJob(FragmentJob::ptr &job)
{
  // work on job and create result
  LOG(2, "DEBUG: Beginning to work on " << job->getId() << ".");
  FragmentResult::ptr result = job->Work();
  LOG(2, "DEBUG: Setting result " << result->getId() << ".");

  AsyncOperation *submitOp = new SubmitResultOperation(connection_, MyAddress, thread_remover, failed);
  static_cast<SubmitResultOperation *>(submitOp)->setResult(result);

  // submit result
  LOG(2, "DEBUG: Putting send result operation into queue ...");
  OpQueue.push_back(submitOp, ServerAddress);
}

/** Wrapper function to allow use as signal handler.
 *
 * We remove us from server's pool and then just call \sa PoolWorker::shutdown().
 *
 * @param sig signal received
 */
void PoolWorker::shutdown(int sig)
{
  LOG(1, "INFO: Shutting down due to signal "+toString(sig)+".");

  shutdown();
}

/** Helper function to shutdown the worker properly.
 *
 * Note that we will use RemoveFromPoolOperation to unlist from server's pool.
 */
void PoolWorker::shutdown()
{
  // remove us from pool
  boost::function<void ()> closingdown = boost::bind(&PoolWorker::finish, this);
  AsyncOperation *removeOp = new RemoveFromPoolOperation(connection_, MyAddress, closingdown, failed);
  LOG(2, "DEBUG: Putting remove from pool operation into queue ...");
  OpQueue.push_back(removeOp, ServerAddress);
  // block queue such that io_service may stop
  OpQueue.block();
}

/** Helper function to close down listener and stop service.
 *
 * This is called after we have been removed from server's pool
 * We stop the io_service via its callback handler in case of success.
 */
void PoolWorker::finish()
{
  // somehow stop listener
  PoolListener.closeSocket();

  // finally, stop io_service
  io_service.stop();
}

